@@ -12,4 +12,4 @@ settings.register_profile("fast", max_examples=5, deadline=5000)
 settings.register_profile("slow", max_examples=20, deadline=5000)
 
 
-pytest_plugins = ["swh.scheduler.pytest_plugin"]
+pytest_plugins = ["swh.scheduler.pytest_plugin", "swh.core.db.pytest_plugin"]
