@@ -6,10 +6,10 @@
 insert into dbversion (version, release, description)
        values (27, now(), 'Work In Progress');
 
-delete function swh_scheduler_peek_ready_tasks (text, timestamptz, bigint, bigint);
-delete function swh_scheduler_peek_priority_tasks (text, timestamptz, bigint);
+drop function swh_scheduler_peek_ready_tasks (text, timestamptz, bigint, bigint);
+drop function swh_scheduler_peek_priority_tasks (text, timestamptz, bigint);
 -- delete old signature function
-delete function swh_scheduler_grab_ready_tasks (text, timestamptz, bigint, bigint);
+drop function swh_scheduler_grab_ready_tasks (text, timestamptz, bigint, bigint);
 
 create or replace function swh_scheduler_grab_ready_tasks (task_type text, ts timestamptz default now(),
                                                            num_tasks bigint default NULL)
