@@ -6,6 +6,7 @@
 __all__ = [
     "SchedulerException",
     "StaleData",
+    "UnknownPolicy",
 ]
 
 
@@ -15,3 +16,7 @@ class SchedulerException(Exception):
 
 class StaleData(SchedulerException):
     pass
+
+
+class UnknownPolicy(SchedulerException):
+    pass
