@@ -24,7 +24,7 @@ TASK_NAMES = ["ping", "multiping", "add", "error", "echo"]
 
 postgresql_scheduler = postgresql_fact(
     "postgresql_proc",
-    db_name="scheduler",
+    dbname="scheduler",
     dump_files=os.path.join(SQL_DIR, "*.sql"),
     no_truncate_tables={"dbversion", "priority_ratio"},
 )
@@ -69,6 +69,7 @@ def swh_scheduler_celery_app():
         task_cls="swh.scheduler.task:SWHTask",
         config={
             "accept_content": ["application/x-msgpack", "application/json"],
+            "broker_url": "memory://guest@localhost//",
             "task_serializer": "msgpack",
             "result_serializer": "json",
         },
