@@ -6,6 +6,7 @@
 from collections import defaultdict
 import copy
 import datetime
+from datetime import timedelta
 import inspect
 import random
 from typing import Any, Dict, List, Optional, Tuple
@@ -18,7 +19,12 @@ import pytest
 from swh.model.hashutil import hash_to_bytes
 from swh.scheduler.exc import SchedulerException, StaleData, UnknownPolicy
 from swh.scheduler.interface import ListedOriginPageToken, SchedulerInterface
-from swh.scheduler.model import ListedOrigin, OriginVisitStats, SchedulerMetrics
+from swh.scheduler.model import (
+    LastVisitStatus,
+    ListedOrigin,
+    OriginVisitStats,
+    SchedulerMetrics,
+)
 from swh.scheduler.utils import utcnow
 
 from .common import (
@@ -29,7 +35,7 @@ from .common import (
     tasks_with_priority_from_template,
 )
 
-ONEDAY = datetime.timedelta(days=1)
+ONEDAY = timedelta(days=1)
 
 NUM_PRIORITY_TASKS = {None: 100, "high": 60, "normal": 30, "low": 20}
 
@@ -760,11 +766,20 @@ class TestScheduler:
 
         return visit_type, recorded_origins
 
-    def _check_grab_next_visit(self, swh_scheduler, visit_type, policy, expected):
-        """Calls grab_next_visits with the passed policy, and check that all
-        the origins returned are the expected ones (in the same order), and
-        that no extra origins are returned. Also checks the origin visits have
-        been marked as scheduled, and are only re-scheduled a week later"""
+    def _check_grab_next_visit_basic(
+        self, swh_scheduler, visit_type, policy, expected, **kwargs
+    ):
+        """Calls grab_next_visits with the passed policy, and check that:
+
+         - all the origins returned are the expected ones (in the same order)
+         - no extra origins are returned
+         - the last_scheduled field has been set properly.
+
+        Pass the extra keyword arguments to the calls to grab_next_visits.
+
+        Returns a timestamp greater than all `last_scheduled` values for the grabbed
+        visits.
+        """
         assert len(expected) != 0
 
         before = utcnow()
@@ -773,6 +788,7 @@ class TestScheduler:
             # Request one more than expected to check that no extra origin is returned
             count=len(expected) + 1,
             policy=policy,
+            **kwargs,
         )
         after = utcnow()
 
@@ -787,16 +803,29 @@ class TestScheduler:
 
         # They should not be scheduled again
         ret = swh_scheduler.grab_next_visits(
-            visit_type=visit_type, count=len(expected) + 1, policy=policy,
+            visit_type=visit_type, count=len(expected) + 1, policy=policy, **kwargs
         )
         assert ret == [], "grab_next_visits returned already-scheduled origins"
 
-        # But a week later, they should
+        return after
+
+    def _check_grab_next_visit(
+        self, swh_scheduler, visit_type, policy, expected, **kwargs
+    ):
+        """Run the same check as _check_grab_next_visit_basic, but also checks the
+        origin visits have been marked as scheduled, and are only re-scheduled a
+        week later
+        """
+
+        after = self._check_grab_next_visit_basic(
+            swh_scheduler, visit_type, policy, expected, **kwargs
+        )
+        # But a week, later, they should
         ret = swh_scheduler.grab_next_visits(
             visit_type=visit_type,
             count=len(expected) + 1,
             policy=policy,
-            timestamp=after + datetime.timedelta(days=7),
+            timestamp=after + timedelta(days=7),
         )
         # We need to sort them because their 'last_scheduled' field is updated to
         # exactly the same value, so the order is not deterministic
@@ -804,8 +833,8 @@ class TestScheduler:
             expected
         ), "grab_next_visits didn't reschedule visits after a week"
 
-    def test_grab_next_visits_oldest_scheduled_first(
-        self, swh_scheduler, listed_origins_by_type,
+    def _prepare_oldest_scheduled_first_origins(
+        self, swh_scheduler, listed_origins_by_type
     ):
         visit_type, origins = self._grab_next_visits_setup(
             swh_scheduler, listed_origins_by_type
@@ -818,11 +847,9 @@ class TestScheduler:
                 url=origin.url,
                 visit_type=origin.visit_type,
                 last_snapshot=None,
-                last_eventful=None,
-                last_uneventful=None,
-                last_failed=None,
-                last_notfound=None,
-                last_scheduled=base_date - datetime.timedelta(seconds=i),
+                last_successful=None,
+                last_visit=None,
+                last_scheduled=base_date - timedelta(seconds=i),
             )
             for i, origin in enumerate(origins[1:])
         ]
@@ -832,6 +859,14 @@ class TestScheduler:
         # as well as those with the oldest values (i.e. the last ones), in order.
         expected = [origins[0]] + origins[1:][::-1]
 
+        return visit_type, origins, expected
+
+    def test_grab_next_visits_oldest_scheduled_first(
+        self, swh_scheduler, listed_origins_by_type,
+    ):
+        visit_type, origins, expected = self._prepare_oldest_scheduled_first_origins(
+            swh_scheduler, listed_origins_by_type
+        )
         self._check_grab_next_visit(
             swh_scheduler,
             visit_type=visit_type,
@@ -839,6 +874,75 @@ class TestScheduler:
             expected=expected,
         )
 
+    @pytest.mark.parametrize("which_cooldown", ("scheduled", "failed", "not_found"))
+    @pytest.mark.parametrize("cooldown", (7, 15))
+    def test_grab_next_visits_cooldowns(
+        self, swh_scheduler, listed_origins_by_type, which_cooldown, cooldown,
+    ):
+        visit_type, origins, expected = self._prepare_oldest_scheduled_first_origins(
+            swh_scheduler, listed_origins_by_type
+        )
+        after = self._check_grab_next_visit_basic(
+            swh_scheduler,
+            visit_type=visit_type,
+            policy="oldest_scheduled_first",
+            expected=expected,
+        )
+
+        # Mark all the visits as scheduled, failed or notfound on the `after` timestamp
+        ovs_args = {
+            "last_visit": None,
+            "last_visit_status": None,
+            "last_scheduled": None,
+        }
+        if which_cooldown == "scheduled":
+            ovs_args["last_scheduled"] = after
+        else:
+            ovs_args["last_visit"] = after
+            ovs_args["last_visit_status"] = LastVisitStatus(which_cooldown)
+
+        visit_stats = [
+            OriginVisitStats(
+                url=origin.url,
+                visit_type=origin.visit_type,
+                last_snapshot=None,
+                last_successful=None,
+                **ovs_args,
+            )
+            for i, origin in enumerate(origins)
+        ]
+        swh_scheduler.origin_visit_stats_upsert(visit_stats)
+
+        cooldown_td = timedelta(days=cooldown)
+        cooldown_args = {
+            "scheduled_cooldown": None,
+            "failed_cooldown": None,
+            "not_found_cooldown": None,
+        }
+        cooldown_args[f"{which_cooldown}_cooldown"] = cooldown_td
+
+        ret = swh_scheduler.grab_next_visits(
+            visit_type=visit_type,
+            count=len(expected) + 1,
+            policy="oldest_scheduled_first",
+            timestamp=after + cooldown_td - timedelta(seconds=1),
+            **cooldown_args,
+        )
+
+        assert ret == [], f"{which_cooldown}_cooldown ignored"
+
+        ret = swh_scheduler.grab_next_visits(
+            visit_type=visit_type,
+            count=len(expected) + 1,
+            policy="oldest_scheduled_first",
+            timestamp=after + cooldown_td + timedelta(seconds=1),
+            **cooldown_args,
+        )
+
+        assert sorted(ret) == sorted(
+            expected
+        ), "grab_next_visits didn't reschedule visits after the configured cooldown"
+
     def test_grab_next_visits_never_visited_oldest_update_first(
         self, swh_scheduler, listed_origins_by_type,
     ):
@@ -849,7 +953,7 @@ class TestScheduler:
         # Update known origins with a `last_update` field that we control
         base_date = datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
         updated_origins = [
-            attr.evolve(origin, last_update=base_date - datetime.timedelta(seconds=i))
+            attr.evolve(origin, last_update=base_date - timedelta(seconds=i))
             for i, origin in enumerate(origins)
         ]
         updated_origins = swh_scheduler.record_listed_origins(updated_origins)
@@ -875,7 +979,7 @@ class TestScheduler:
         # Update known origins with a `last_update` field that we control
         base_date = datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
         updated_origins = [
-            attr.evolve(origin, last_update=base_date - datetime.timedelta(seconds=i))
+            attr.evolve(origin, last_update=base_date - timedelta(seconds=i))
             for i, origin in enumerate(origins)
         ]
         updated_origins = swh_scheduler.record_listed_origins(updated_origins)
@@ -890,10 +994,8 @@ class TestScheduler:
                 url=origin.url,
                 visit_type=origin.visit_type,
                 last_snapshot=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
-                last_eventful=visit_date,
-                last_uneventful=None,
-                last_failed=None,
-                last_notfound=None,
+                last_successful=visit_date,
+                last_visit=visit_date,
             )
             for origin in visited_origins
         ]
@@ -930,6 +1032,131 @@ class TestScheduler:
 
         assert len(ret) == 5
 
+    def test_grab_next_visits_no_last_update_nor_visit_stats(
+        self, swh_scheduler, listed_origins_by_type
+    ):
+        """grab_next_visits should retrieve tasks without last update (nor visit stats)
+
+        """
+        visit_type = next(iter(listed_origins_by_type))
+
+        origins = []
+        for origin in listed_origins_by_type[visit_type]:
+            origins.append(
+                attr.evolve(origin, last_update=None)
+            )  # void the last update so we are in the relevant context
+
+        assert len(origins) > 0
+
+        swh_scheduler.record_listed_origins(origins)
+
+        # Initially, we have no global queue position
+        actual_state = swh_scheduler.visit_scheduler_queue_position_get()
+        assert actual_state == {}
+
+        # nor any visit statuses
+        actual_visit_stats = swh_scheduler.origin_visit_stats_get(
+            (o.url, o.visit_type) for o in origins
+        )
+        assert len(actual_visit_stats) == 0
+
+        # Grab some new visits
+        next_visits = swh_scheduler.grab_next_visits(
+            visit_type, count=len(origins), policy="origins_without_last_update",
+        )
+        # we do have the one without any last update
+        assert len(next_visits) == len(origins)
+
+        # Now the global state got updated
+        actual_state = swh_scheduler.visit_scheduler_queue_position_get()
+        assert actual_state[visit_type] is not None
+
+        actual_visit_stats = swh_scheduler.origin_visit_stats_get(
+            (o.url, o.visit_type) for o in next_visits
+        )
+
+        # Visit stats got algo created
+        assert len(actual_visit_stats) == len(origins)
+
+    def test_grab_next_visits_no_last_update_with_visit_stats(
+        self, swh_scheduler, listed_origins_by_type
+    ):
+        """grab_next_visits should retrieve tasks without last update"""
+        visit_type = next(iter(listed_origins_by_type))
+
+        origins = []
+        for origin in listed_origins_by_type[visit_type]:
+            origins.append(
+                attr.evolve(origin, last_update=None)
+            )  # void the last update so we are in the relevant context
+
+        assert len(origins) > 0
+
+        swh_scheduler.record_listed_origins(origins)
+
+        # Initially, we have no global queue position
+        actual_state = swh_scheduler.visit_scheduler_queue_position_get()
+        assert actual_state == {}
+
+        date_now = utcnow()
+        # Simulate some of those origins have associated visit stats (some with an
+        # existing queue position and some without any)
+        visit_stats = (
+            [
+                OriginVisitStats(
+                    url=origin.url,
+                    visit_type=origin.visit_type,
+                    last_successful=utcnow(),
+                    last_visit=utcnow(),
+                    next_visit_queue_position=date_now
+                    + timedelta(days=random.uniform(-10, 1)),
+                )
+                for origin in origins[:100]
+            ]
+            + [
+                OriginVisitStats(
+                    url=origin.url,
+                    visit_type=origin.visit_type,
+                    last_successful=utcnow(),
+                    last_visit=utcnow(),
+                    next_visit_queue_position=date_now
+                    + timedelta(days=random.uniform(1, 10)),  # definitely > now()
+                )
+                for origin in origins[100:150]
+            ]
+            + [
+                OriginVisitStats(
+                    url=origin.url,
+                    visit_type=origin.visit_type,
+                    last_successful=utcnow(),
+                    last_visit=utcnow(),
+                )
+                for origin in origins[150:]
+            ]
+        )
+
+        swh_scheduler.origin_visit_stats_upsert(visit_stats)
+
+        # Grab next visits
+        actual_visits = swh_scheduler.grab_next_visits(
+            visit_type, count=len(origins), policy="origins_without_last_update",
+        )
+        assert len(actual_visits) == len(origins)
+
+        actual_visit_stats = swh_scheduler.origin_visit_stats_get(
+            (o.url, o.visit_type) for o in actual_visits
+        )
+        assert len(actual_visit_stats) == len(origins)
+
+        actual_state = swh_scheduler.visit_scheduler_queue_position_get()
+        assert actual_state == {
+            visit_type: max(
+                s.next_visit_queue_position
+                for s in actual_visit_stats
+                if s.next_visit_queue_position is not None
+            )
+        }
+
     def test_grab_next_visits_unknown_policy(self, swh_scheduler):
         unknown_policy = "non_existing_policy"
         NUM_RESULTS = 5
@@ -950,10 +1177,8 @@ class TestScheduler:
             OriginVisitStats(
                 url=f"https://example.com/origin-{i:03d}",
                 visit_type="git",
-                last_eventful=utcnow(),
-                last_uneventful=None,
-                last_failed=None,
-                last_notfound=None,
+                last_successful=utcnow(),
+                last_visit=utcnow(),
             )
             for i in range(
                 page_size + 1
@@ -975,10 +1200,8 @@ class TestScheduler:
         visit_stats = OriginVisitStats(
             url=url,
             visit_type="git",
-            last_eventful=eventful_date,
-            last_uneventful=None,
-            last_failed=None,
-            last_notfound=None,
+            last_successful=eventful_date,
+            last_visit=eventful_date,
         )
         swh_scheduler.origin_visit_stats_upsert([visit_stats])
         swh_scheduler.origin_visit_stats_upsert([visit_stats])
@@ -986,14 +1209,9 @@ class TestScheduler:
         assert swh_scheduler.origin_visit_stats_get([(url, "git")]) == [visit_stats]
         assert swh_scheduler.origin_visit_stats_get([(url, "svn")]) == []
 
-        uneventful_date = utcnow()
+        new_visit_date = utcnow()
         visit_stats = OriginVisitStats(
-            url=url,
-            visit_type="git",
-            last_eventful=None,
-            last_uneventful=uneventful_date,
-            last_failed=None,
-            last_notfound=None,
+            url=url, visit_type="git", last_successful=None, last_visit=new_visit_date,
         )
         swh_scheduler.origin_visit_stats_upsert([visit_stats])
 
@@ -1002,38 +1220,12 @@ class TestScheduler:
         expected_visit_stats = OriginVisitStats(
             url=url,
             visit_type="git",
-            last_eventful=eventful_date,
-            last_uneventful=uneventful_date,
-            last_failed=None,
-            last_notfound=None,
+            last_successful=eventful_date,
+            last_visit=new_visit_date,
         )
 
         assert uneventful_visits == [expected_visit_stats]
 
-        failed_date = utcnow()
-        visit_stats = OriginVisitStats(
-            url=url,
-            visit_type="git",
-            last_eventful=None,
-            last_uneventful=None,
-            last_failed=failed_date,
-            last_notfound=None,
-        )
-        swh_scheduler.origin_visit_stats_upsert([visit_stats])
-
-        failed_visits = swh_scheduler.origin_visit_stats_get([(url, "git")])
-
-        expected_visit_stats = OriginVisitStats(
-            url=url,
-            visit_type="git",
-            last_eventful=eventful_date,
-            last_uneventful=uneventful_date,
-            last_failed=failed_date,
-            last_notfound=None,
-        )
-
-        assert failed_visits == [expected_visit_stats]
-
     def test_origin_visit_stats_upsert_with_snapshot(self, swh_scheduler) -> None:
         eventful_date = utcnow()
         url = "https://github.com/666/test"
@@ -1041,10 +1233,7 @@ class TestScheduler:
         visit_stats = OriginVisitStats(
             url=url,
             visit_type="git",
-            last_eventful=eventful_date,
-            last_uneventful=None,
-            last_failed=None,
-            last_notfound=None,
+            last_successful=eventful_date,
             last_snapshot=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
         )
         swh_scheduler.origin_visit_stats_upsert([visit_stats])
@@ -1058,19 +1247,13 @@ class TestScheduler:
             OriginVisitStats(
                 url="foo",
                 visit_type="git",
-                last_eventful=utcnow(),
-                last_uneventful=None,
-                last_failed=None,
-                last_notfound=None,
+                last_successful=utcnow(),
                 last_snapshot=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
             ),
             OriginVisitStats(
                 url="bar",
                 visit_type="git",
-                last_eventful=None,
-                last_uneventful=utcnow(),
-                last_notfound=None,
-                last_failed=None,
+                last_visit=utcnow(),
                 last_snapshot=hash_to_bytes("fffcc0710eb6cf9efd5b920a8453e1e07157bfff"),
             ),
         ]
@@ -1092,24 +1275,38 @@ class TestScheduler:
                     OriginVisitStats(
                         url="foo",
                         visit_type="git",
-                        last_eventful=None,
-                        last_uneventful=utcnow(),
-                        last_notfound=None,
-                        last_failed=None,
-                        last_snapshot=None,
+                        last_successful=None,
+                        last_visit=utcnow(),
                     ),
                     OriginVisitStats(
                         url="foo",
                         visit_type="git",
-                        last_eventful=None,
-                        last_uneventful=utcnow(),
-                        last_notfound=None,
-                        last_failed=None,
-                        last_snapshot=None,
+                        last_successful=utcnow(),
+                        last_visit=None,
                     ),
                 ]
             )
 
+    def test_visit_scheduler_queue_position(
+        self, swh_scheduler, listed_origins
+    ) -> None:
+        result = swh_scheduler.visit_scheduler_queue_position_get()
+        assert result == {}
+
+        expected_result = {}
+        visit_types = set()
+        for origin in listed_origins:
+            visit_type = origin.visit_type
+            if visit_type in visit_types:
+                continue
+            visit_types.add(visit_type)
+            position = utcnow()
+            swh_scheduler.visit_scheduler_queue_position_set(visit_type, position)
+            expected_result[visit_type] = position
+
+        result = swh_scheduler.visit_scheduler_queue_position_get()
+        assert result == expected_result
+
     def test_metrics_origins_known(self, swh_scheduler, listed_origins):
         swh_scheduler.record_listed_origins(listed_origins)
 
@@ -1141,10 +1338,7 @@ class TestScheduler:
                 OriginVisitStats(
                     url=visited_origin.url,
                     visit_type=visited_origin.visit_type,
-                    last_eventful=utcnow(),
-                    last_uneventful=None,
-                    last_failed=None,
-                    last_notfound=None,
+                    last_successful=utcnow(),
                     last_snapshot=hash_to_bytes(
                         "d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"
                     ),
@@ -1173,11 +1367,7 @@ class TestScheduler:
                 OriginVisitStats(
                     url=visited_origin.url,
                     visit_type=visited_origin.visit_type,
-                    last_eventful=visited_origin.last_update
-                    - datetime.timedelta(days=1),
-                    last_uneventful=None,
-                    last_failed=None,
-                    last_notfound=None,
+                    last_successful=visited_origin.last_update - timedelta(days=1),
                     last_snapshot=hash_to_bytes(
                         "d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"
                     ),
@@ -1220,7 +1410,7 @@ class TestScheduler:
         ret = swh_scheduler.update_metrics(timestamp=ts)
         assert all(metric.last_update == ts for metric in ret)
 
-        second_ts = ts + datetime.timedelta(seconds=1)
+        second_ts = ts + timedelta(seconds=1)
         ret = swh_scheduler.update_metrics(timestamp=second_ts)
         assert all(metric.last_update == second_ts for metric in ret)
 
